#undef __NO_VERSION__

#include "odm.h"
#include "oim.h"
#include "nsmtypes.h"
#include "nsm.h"

int ODMOpen (struct inode *inode, struct file *filp);
int ODMClose (struct inode *inode, struct file *filp);
int ODMIoctl(struct inode *inode, struct file *filp, unsigned int cmd, 
	unsigned long arg );

int init_module(void);
void cleanup_module(void);
void *ODMMalloc(UINT nSize,UINT nFlags);
void ODMFree(VOID  *ptr,UINT nSize,UINT nFlags);
void ODMCopy(VOID *Source,VOID *Destination,UINT Length);
int  ODMMemCmp(VOID *Sorce1,VOID *Sorce2,UINT Length);
void ODMMemSet(VOID *Destination,UCHAR Fill,ULONG Length);
void ODMMallocContiguous(AddressStruct *pAddrStruct,UINT nFlags);
void ODMFreeContiguousMemory(VOID *ptr,UINT nFlags);
//Added by jaggu
void ODMSetMacAddr(AdapterContext *);
int ODMGetPortNum(AdapterContext *pAdapter); 
void ODMSetdev_aggr(AdapterContext *dest, AdapterContext *src); 
void ODMLacpStart(USHORT nPort);

int ascitoint(char *port);

//Added by jaggu 
int compare(const char *devname, const char *interface); 
int strlenth(char *output);
void strreverse(char *output);
void stringcpy(char *dest, char *src); 
char *format(int);

extern USHORT LacpConfig(PortInfo_t *pPort_Info);
extern void *GlobalLock;
extern	lacpPort_t 	*LACPPortp[MAX_PORTS];
extern	PortInfo_t	*DeferPortInfoList[MAX_PORTS];
//bjaggu
extern  UINT 	AdapterCount;
extern	AdapterContext *AdapterArray[MAX_PORTS];

extern void AggrReshuffle(lacpAggr_t *pAggr);
extern void LacpWaitWhileTimer(void *arg);

UINT PortsInAggr=2;

#if defined(MODULE)

MODULE_DESCRIPTION("PLANET ENW-9602T Gigabit NIC Server driver");

#endif

struct file_operations ODM_fops = {ioctl: ODMIoctl, open:  ODMOpen};
									
int ODM_major = 0;

void ODMWakeup(void *arg)
{
	return;
}

void ODMDelay(ULONG time)
{
	mdelay(time);
	return;
}

void ODMAdapterOpen(void)	
{
	return;
}

void ODMPostEvent(lacpPort_t *pPort,USHORT Event,HsmRCB *Rcb)
{
	return;
}

void ODMSignalProcessPackets(void)
{
	return;
}

UINT ODMProcessPacketsRunning(void)
{
	return TRUE;
}

UCHAR ODMGetSlot(AdapterContext *pAdapterContext)
{
	return pAdapterContext->Slot_Number;
}

UCHAR ODMGetIOBus(AdapterContext *pAdapterContext)
{
	return pAdapterContext->Bus_Number;
}

ULONG ODMGetPresentTimeInTicks(void)
{
	return jiffies;
}

void ODMInitializeLock(void **ptr,int nFlags)
{
	*ptr = kmalloc(sizeof(spinlock_t), GFP_KERNEL);
	
	if(*ptr == NULL)
		return;
	
	spin_lock_init((spinlock_t *)(*ptr));	
	
}

void ODMUnInitializeLock(void *ptr,int nFlags)
{
	kfree(ptr);
}

PRIV_LEVEL ODMAcquireLock(void *ptr,int nFlags,PRIV_LEVEL NewPrivLevel)
{
	spin_lock((spinlock_t *)ptr);
	return NewPrivLevel;
}

void ODMReleaseLock(void *ptr,int nFlags,PRIV_LEVEL PrivLevel)
{
	spin_unlock((spinlock_t *)ptr);
}

VOID timer_func(ULONG ptr)
{
	lacpPort_t *pPort;
	lacpAggr_t *pAggr;
	TimerContext_t *pTimerContext = (TimerContext_t *)ptr;

	if(ptr == 0)
		return;

	//printk("\nTimer Type %x", pTimerContext->TimerType);

	if(pTimerContext->TimerType == WAIT_WHILE_TIMER)
	{
		pPort = (lacpPort_t *)(pTimerContext->pContext);
		LacpWaitWhileTimer(pPort);
		pPort->TimerType  &= ~WAIT_WHILE_TIMER;
		kfree((void *)pTimerContext);
	}
	if(pTimerContext->TimerType == AGGR_RESHUFFLE_TIMER)
	{
		//DbgPrint("\nCalling AggrReshuffle");
		pAggr = (lacpAggr_t *)(pTimerContext->pContext);
		AggrReshuffle(pAggr);
		pAggr->TimerType  &= ~AGGR_RESHUFFLE_TIMER;
		kfree((void *)pTimerContext);
	}

	return;
}

void *ODMTimeout(void (*pFunc)(),void *ptr,ULONG nTicks,int nFlags)
{
	lacpPort_t			*pPort;
	lacpAggr_t			*pAggr;
	TimerContext_t		*pTimerContext;	
	struct timer_list 	*pTimer;

	//DbgPrint1("\nOT: Flags %x", nFlags); 

	pTimerContext = (TimerContext_t *)kmalloc(sizeof(TimerContext_t), 
		GFP_ATOMIC);
	if(pTimerContext == NULL)
		return NULL;

	switch(nFlags)
	{
		case WAIT_WHILE_TIMER:
			pPort = (lacpPort_t *)ptr;
			pTimerContext->TimerType = WAIT_WHILE_TIMER;		 
			pPort->TimerType |= WAIT_WHILE_TIMER;

			/* This can perhaps be replaced with the WaitWhileTimerID */
			if(pPort->WaitWhileTimerID == 0)
			{
				//printk("\nWaitWhileTimerID is NULL");
				pTimer = (struct timer_list *)
					kmalloc(sizeof(struct timer_list), GFP_ATOMIC);
	
				if(pTimer == NULL)
				{
					kfree(pTimerContext);
					return NULL;
				}
				init_timer(pTimer);
			}
			else
				pTimer = (struct timer_list *) pPort->WaitWhileTimerID;
			break;

		case AGGR_RESHUFFLE_TIMER:
			pAggr = (lacpAggr_t *)ptr;
			pTimerContext->TimerType = AGGR_RESHUFFLE_TIMER;
			pAggr->TimerType |= AGGR_RESHUFFLE_TIMER;
			
			if(pAggr->ReshuffleTimerID == 0)
			{
				//printk("\nReshuffleTimerID is NULL");
				pTimer = (struct timer_list *)
					kmalloc(sizeof(struct timer_list), GFP_ATOMIC);
	
				if(pTimer == NULL)
				{
					kfree(pTimerContext);
					return NULL;
				}
				init_timer(pTimer);
			}
			else 
				pTimer = (struct timer_list *)pAggr->ReshuffleTimerID;
			break;
		
		default:
			kfree(pTimerContext);
			return NULL;
	}

	pTimerContext->pContext = ptr;

	pTimer->expires = jiffies + nTicks;
	pTimer->data = (ULONG)pTimerContext;
	pTimer->function = &timer_func;	/* timer handler */
	add_timer(pTimer);

	return ((void *)pTimer); 	
}

void ODMUntimeout(void *ptr,void *pContext)
{
	if(((struct timer_list *)ptr)->data)
	{
		kfree((void *) (((struct timer_list *)ptr)->data));
		((struct timer_list *)ptr)->data = 0;
	}
	del_timer((struct timer_list *)ptr);
	kfree(ptr);
}


VOID *ODMMalloc(UINT nSize, UINT nFlags)
{
	VOID *ptr;
	
	ptr = kmalloc(nSize, GFP_KERNEL);
	
	if(ptr == NULL)
		return NULL;
	
	memset(ptr,0,nSize);
	return ptr;
}

VOID ODMFree(VOID *ptr,UINT nSize, UINT nFlags)
{
	kfree(ptr);
}

VOID ODMCopy(VOID *Source,VOID *Destination,UINT Length)
{
	memcpy(Destination,Source,Length);
}

int  ODMMemCmp(VOID *Source1,VOID *Source2,UINT Length)
{
	if((memcmp(Source1, Source2, Length)) != 0 )
		return FALSE;
	return TRUE;
}

VOID ODMMemSet(VOID *Destination,UCHAR Fill,ULONG Length)
{
	memset(Destination,Fill,Length);
}

VOID ODMMallocContiguous(AddressStruct *pAddrStruct,UINT nFlags)
{
	pAddrStruct->VirtAddr = kmalloc (pAddrStruct->BufSize, 
		GFP_KERNEL | GFP_ATOMIC);
	
	if(pAddrStruct->VirtAddr == NULL)
	{
		printk("\nAllocation of contiguous memory failed");
		
		pAddrStruct->PhysAddr = NULL;

		return;

	}
	
	memset(pAddrStruct->VirtAddr , 0, pAddrStruct->BufSize);	

	pAddrStruct->pCookie = pAddrStruct->VirtAddr;

	pAddrStruct->PhysAddr = 
		(UCHAR *)virt_to_bus ((VOID *)pAddrStruct->VirtAddr);

}


VOID  ODMFreeContiguousMemory(VOID *ptr, UINT nFlags)
{
	AddressStruct *pAddrStruct = (AddressStruct *)ptr;
	kfree(pAddrStruct->VirtAddr);
	pAddrStruct->VirtAddr = 0;
	pAddrStruct->PhysAddr = 0;
	pAddrStruct->pCookie = 0;
}

int ODMOpen (struct inode *inode, struct file *filp)                            
{
    printk(KERN_INFO "lacp: The pseudo driver Opened");
    return 0;
}

int ODMClose (struct inode *inode, struct file *filp)
{
    printk(KERN_INFO "lacp: The pseudo driver Closed");
    return 0;
}                                                                             

int
init_module (void)
{
    int result;

#ifdef _DUMP
	struct module 	*mlist = &__this_module;
	printk("LACP module %s is loaded at %p\n\n",mlist->name,
		&((char *)mlist)[mlist->size_of_struct]);
#endif

    result = register_chrdev(ODM_major, "lacp", &ODM_fops);
    if(result < 0)
    {
        printk(KERN_WARNING "lacp: can't get major %d\n", ODM_major);
        return result;
    }
    if(ODM_major == 0)
        ODM_major = result;

    printk(KERN_INFO "lacp: major number %d\n", ODM_major);
    return 0;
}                                                                              

void cleanup_module (void)
{
    unregister_chrdev(ODM_major, "lacp");
    printk(KERN_INFO "lacp : unloaded\n");
}                                                                              


int ODMIoctl( struct inode *inode, struct file *filp, unsigned int cmd, 
	unsigned long arg )
{
	PortInfo_t  	*pPortInfo;
	PortInfo_t	PortInfo;
	char		*port, *AggrId;
	NsmContext	*pNsmContext;
	AdapterContext	*pAdapter;
	int		i;

//Added by jaggu
	switch (cmd)
	{
		case IOCTL_LACP_CONFIG:
		{
			/* Copies data bytes from user space to kernel space */
			memset(&PortInfo,0,sizeof(PortInfo_t));
			if (copy_from_user(&PortInfo, (PortInfo_t *)arg, sizeof(PortInfo_t)))
				return -EFAULT;
			
			pPortInfo = &PortInfo;
/* 
printk("\nLacp Driver : LACP	 = %d\n",pPortInfo->LACP);
printk("\nLacp Driver : GEC	 = %d\n",pPortInfo->GEC);
printk("\nLacp Driver : LACPActivity= %d\n",pPortInfo->LACPActivity);
printk("\nLacp Driver : Aggregator Port= %d\n",pPortInfo->AggregatorPort);
printk("\nLacp Driver : Port Num = %d\n",pPortInfo->Port);
printk("\nLacp Driver : Key	 = %d\n",pPortInfo->Key);
printk("\nLacp Driver : AggregatorId= %d\n",pPortInfo->AggregatorIdentifier);
*/

			port = format(pPortInfo->Port);
			printk("\nPortNumber  : LacpDriver = %s\n", port);

			for(i=0; i<MAX_PORTS; i++) {
				if(AdapterArray[i] == NULL)
					continue; 
		
				pAdapter = AdapterArray[i];
				pNsmContext = pAdapter->pNsmContext;
				if(!compare(pNsmContext->dev->name, port)) {
					pPortInfo->Port = i;
					break;
				}
				else
					continue;
			}

			if(i==MAX_PORTS)
			{
				ODMFree(port,INTERFACE_SIZE,0);
				printk(KERN_INFO "Port not present in the system\n");
				return -EFAULT;
			}
			
			ODMFree(port,INTERFACE_SIZE,0);

			AggrId = format(pPortInfo->AggregatorIdentifier);
			printk("AggregatorId: LacpDriver = %s\n\n", AggrId);

			for(i=0; i<MAX_PORTS ; i++) {
				if(AdapterArray[i] == NULL)
					continue; 
				pAdapter = AdapterArray[i];
				pNsmContext = pAdapter->pNsmContext;
				if(!compare(pNsmContext->dev->name, AggrId)) {
					pPortInfo->AggregatorIdentifier= i;
					break;
				}
				else
					continue;
			}

			if(i==MAX_PORTS)
			{
				ODMFree(AggrId,INTERFACE_SIZE,0);
				printk(KERN_INFO "AggregatorId not present in the system\n");
				return -EFAULT;
			}
			ODMFree(AggrId,INTERFACE_SIZE,0);

			LacpConfig(pPortInfo);
		}
		break;

		default:
				return -EINVAL;
	}		
    return 0;
}                                                                             

inline VOID PrintContents(UCHAR *start, UINT len)
{
	UCHAR	*ptr = start;
	UINT 	i;
	printk("The contents of the location :\n");
	for(i=0;i<len;i++)
	{
		printk(": %x :"	,*ptr);			
		ptr++;
	}
}


//Following modules Added by jaggu
char *
format(int num) {
	int	i,j=0;
	char	temp;
	char	output[INTERFACE_SIZE];
	char	*interface;
	
	if((interface= ODMMalloc(INTERFACE_SIZE,ZERO_MEMORY)) == NULL)
		return NULL;
	
	stringcpy(interface, "eth");
	
	if(num < 0)
		return NULL;
	else if(num!=0) {
		for(i=0; num != 0; i++) {
			temp = num%10;
			output[i]=temp;
			num=num/10;
		}
		output[i] = (char) 0;
		strreverse(output);
	}else {
		interface[3]='0';
		interface[4]=(char)0;
		return interface;
	}

	while(*(interface+j) != '\0') {
		j++;		
	}
	
	for(i=0; (temp = *(output+i)) != '\0'; i++) {
		interface[j++]=temp+'0';
	}

	interface[j] = (char) 0;

	return interface;
}

int
compare(const char *devname, const char *interface) 
{
	signed char ret;

	while (1) {
		if ((ret = *devname - *interface++) != 0 || !*devname++)
			break;
	}
	return ret;
}

void
strreverse(char *output)
{
	int c,i,j;

	for(i=0, j = strlenth(output)-1; i<j;i++,j--) {
		c = *(output+i);

		*(output+i) = *(output+j);

		*(output+j) = c;
	}

}

int
strlenth(char *output){
	char *p = output;

	while(*output != '\0') {
		output++;		
	}

	return output - p;
}

void
stringcpy(char *dest, char *src) {
if(src)
	while(*dest++ = *src++);
}

void ODMSetMacAddr(AdapterContext *pAdapter)
{
	ODMCopy((VOID *)pAdapter->CurrMacAddr,(VOID *) ((NsmContext *)(pAdapter->pNsmContext))->dev->dev_addr, MAC_ADDR_LEN);
}

int
ODMGetPortNum(AdapterContext *pAdapter) {
	char interface[INTERFACE_SIZE];
	int	portnum;

    stringcpy(interface,((NsmContext *)(pAdapter->pNsmContext))->dev->name);
    portnum = ascitoint(interface+3);

	return portnum;
}

void
ODMSetdev_aggr(AdapterContext *dest, AdapterContext *src) {
	((NsmContext *)dest->pNsmContext)->dev_aggr = ((NsmContext *)(src->pNsmContext))->dev;
}

int
ascitoint(char *port) {
	int i,ret;
        ret=0;
	for(i=0; *(port+i) >= '0' && *(port+i) <= '9' ;i++) {
		ret = ret * 10 + ( *(port+i) - '0');
	}
	return ++ret;
}

void
ODMLacpStart(USHORT nPort)
{
	return;
}
