#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/ioctl.h>
#include "portinfo.h"

void fill_portdata(int,int,int,int,char *,char *,char *); 
void issue_configioctl(void);
void read_from_configfile(char *, FILE *); 

int			lacpfd;
PortInfo_t	PortInfo;
int			ConfPorts[MAX_PORTS];

main(int argc, char **argv) {
    char 	aggrport[2];
	FILE	*confp;
    int		i;

	lacpfd = open(DRIVER_PATH, O_RDWR);
	if (lacpfd < 0) {
		fprintf(stderr,"Failed to open LACP Driver\n");
		exit(1);
	}

	if(argc ==2 ) {
		confp = fopen(argv[1], "r");
		if (confp == NULL) {
			fprintf(stderr,"Failed to open lacpconfig file\n");
			close(lacpfd);
			exit(1);
		}
	}
	else 
	{
		fprintf(stderr,"Usage : lacpstart [CONFIG_FILE_PATH]\n");
		close(lacpfd);
		exit(1);
	}
	for(i=0;i<MAX_PORTS;i++)
		ConfPorts[i]=-1;

	strcpy(aggrport,"y");
	read_from_configfile(aggrport, confp);
	fseek(confp,0,SEEK_SET);
	strcpy(aggrport,"n");
	read_from_configfile(aggrport,confp);

	close(lacpfd);
	close(confp);
}

void
read_from_configfile(char *port_type, FILE *confp) {
	int		Flag,Key,LacpActivity,LacpTimeout,i;
	char 	Port[10],AggregatorPort[5],AggregatorId[10];
	char	ConfData[MAX_PORTS];

	while(fgets(ConfData, 256, confp)) {
		if(ConfData[0] == ';') 
			continue;
		sscanf(ConfData,"%[^:]:%d:%d:%d:%d:%[^:]:%s\n",Port,&Flag,&Key,&LacpActivity,&LacpTimeout,AggregatorPort,AggregatorId);

	 	if(!strcmp(AggregatorPort,port_type)) {  
			// Configure all Aggregator Port 
			fill_portdata(Flag,Key,LacpActivity,LacpTimeout,Port,AggregatorPort,AggregatorId);

			if(PortInfo.AggregatorPort) {
	 			if(strcmp(Port,AggregatorId)) {  
					fprintf(stderr,"Could not configure Port %s : For Aggregator Port AggregatorId should be same as Port\n",Port);
					close(lacpfd);
					close(confp);
					exit(1);
				}
				for(i=0;i<MAX_PORTS;i++)
					if(ConfPorts[i] == -1) {
						ConfPorts[i] = PortInfo.Port;
						break;
					}
			} else {
				for(i=0;i<MAX_PORTS;i++)
					if(ConfPorts[i] == PortInfo.AggregatorIdentifier) {
						break;
					}
					if(i==MAX_PORTS) {
						fprintf(stderr,"Error :Could not configure port %s, Given Aggregator Identifier is not proper\n",Port);
						close(lacpfd);
						close(confp);
						exit(1);
					}
			}
			issue_configioctl();
		}
	}//end of while
}

void
issue_configioctl() {
		// Issue LACP_CONFIG IOCTL to configure Aggregation 
		if( ioctl(lacpfd, IOCTL_LACP_CONFIG, &PortInfo) < 0) {
			fprintf(stderr,"Could not issue IOCTL_LACP_CONFIG\n");
			close(lacpfd);
			exit(1);
		}
}

void
fill_portdata(int Flag,int Key,int LacpActivity,int LacpTimeout, char Port[10], char AggregatorPort[5],char AggregatorId[10]) {

	if((Flag & LACP_FLAG))
		PortInfo.LACP = TRUE;
	else
		PortInfo.LACP = FALSE;

	if((Flag & GEC_FLAG))
		PortInfo.GEC = TRUE;
	else
		PortInfo.GEC = FALSE;

	PortInfo.LACPActivity = LacpActivity;

	if(!strcmp(AggregatorPort,"y"))
		PortInfo.AggregatorPort=TRUE;
	else
		PortInfo.AggregatorPort=FALSE;

	sscanf(Port,"%*[^0-9]%d",&PortInfo.Port);

	PortInfo.Key = Key;

	sscanf(AggregatorId,"%*[^0-9]%d",&PortInfo.AggregatorIdentifier);
		
	if( LacpTimeout == 1)
		PortInfo.LACPTimeout = TRUE;	
	else
		PortInfo.LACPTimeout = FALSE;	
}
